from asyncio import sleep
from asyncio.events import get_event_loop
from asyncio.locks import Event
from datetime import datetime
from functools import wraps
from typing import Any, Dict, Literal, Optional
from logging import getLogger

import aiohttp

from .decorator import strict_literal
from .errors import ERROR_MAPPING, HTTPException, AuthorizeError
from .typing import WidgetStyle, WidgetType

BASE = "https://koreanbots.dev/api/"
VERSION = "v2"

KOREANBOTS_URL = BASE + VERSION

log = getLogger(__name__)


def required(f: Any):
    @wraps(f)
    async def decorator_function(
        self: "KoreanbotsRequester", *args: Any, **kwargs: Any
    ):
        if not self.api_key:
            raise AuthorizeError("This endpoint required koreanbots token.")

        return await f(self, *args, **kwargs)

    return decorator_function


class KoreanbotsRequester:
    """
    Koreanbots의 API를 요청하는 클래스입니다.

    :param api_key:
        KoreanBots의 토큰입니다. 기본값은 None 입니다.
    :type api_key:
        Optional[str], optional

    :param session:
        aiohttp.ClientSession의 클래스입니다. 전달되지 않으면 생성합니다. 기본값은 None 입니다.
    :type session:
        Optional[aiohttp.ClientSession], optional
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        self.session = session
        self.api_key = api_key
        self._global_limit = Event()
        self._global_limit.set()

    # How to close the session if discord.Client is not specified.
    def __del__(self):
        if self.session:
            if not self.session.closed:
                loop = get_event_loop()
                if loop.is_running():
                    loop.create_task(self.session.close())
                else:
                    loop.run_until_complete(self.session.close())

    async def request(
        self,
        method: Literal["GET", "POST"],
        endpoint: str,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """
        Koreanbots의 url을 기반으로 요청합니다.
        레이트리밋을 핸들합니다.

        :param method:
            HTTP 메소드입니다. GET, POST만 사용할 수 있습니다.
        :type method:
            Literal["GET", "POST"]
        :param endpoint:
            요청을 실행할 API 페이지의 주소입니다.
        :type endpoint:
            str

        :raises NotFound:
            요청할 수 없는 페이지입니다.
        :raises BadRequest:
            잘못된 요청입니다.
        :raises Forbidden:
            요청을 할 권한이 없습니다.

        :raises HTTPException:
            응답에 오류가 있습니다.

        :return:
            요청 결과를 반환합니다.
        :rtype:
            Dict[str, Any]
        """

        if not self.session:
            self.session = aiohttp.ClientSession()

        if not self._global_limit.is_set():
            await self._global_limit.wait()

        for _ in range(5):
            async with self.session.request(
                method, KOREANBOTS_URL + endpoint, **kwargs
            ) as response:
                remain_limit = response.headers["x-ratelimit-remaining"]
                if remain_limit == 0 or response.status == 429:
                    reset_limit_timestamp = int(response.headers["x-ratelimit-reset"])
                    reset_limit = datetime.fromtimestamp(reset_limit_timestamp)
                    retry_after = reset_limit - datetime.now()
                    self._global_limit.clear()
                    await sleep(retry_after.total_seconds())
                    self._global_limit.set()
                    continue

                if response.status != 200:
                    if ERROR_MAPPING.get(response.status):
                        raise ERROR_MAPPING[response.status](
                            response.status, await response.json()
                        )
                    else:
                        raise HTTPException(response.status, await response.json())
                return await response.json()

        assert None

    async def get_bot_info(self, bot_id: int) -> Dict[str, Any]:
        """
        주어진 bot_id로 bot의 정보를 반환합니다.

        :param bot_id:
            요청할 bot의 ID를 지정합니다.
        :type bot_id:
            int

        :return:
            요청 결과를 반환합니다.
        :rtype:
            Dict[str, Any]
        """
        return await self.request("GET", f"/bots/{bot_id}")

    @required
    async def post_update_bot_info(self, bot_id: int, **kwargs: int) -> Dict[str, Any]:
        """
        주어진 bot_id로 bot의 정보를 갱신합니다.

        :param bot_id:
            요청할 bot의 ID를 지정합니다.
        :type bot_id:
            int

        :param kwargs:
            갱신할 정보를 지정합니다.
            'servers' 인자와 'shards' 인자 이외의 값이 들어갈경우 무시합니다.
        :type kwargs:
            int

        :raises AuthorizeError:
            api_key가 없거나 유효하지 않은 경우。

        :return:
            요청 결과를 반환합니다.
        :rtype:
            Dict[str, Any]
        """

        return await self.request(
            "POST",
            f"/bots/{bot_id}/stats",
            json={x: kwargs[x] for x in kwargs if x not in ["servers", "shards"]},
            headers={"Authorization": self.api_key},
        )

    @strict_literal("widget_type")
    @strict_literal("style")
    async def get_bot_widget_url(
        self,
        widget_type: WidgetType,
        bot_id: int,
        style: WidgetStyle = "flat",
        scale: float = 1.0,
        icon: bool = False,
    ) -> str:
        """
        주어진 bot_id로 widget의 url을 반환합니다.

        :param widget_type:
            요청할 widget의 타입을 지정합니다.
        :type widget_type:
            WidgetType

        :param bot_id:
            요청할 bot의 ID를 지정합니다.
        :type bot_id:
            int

        :param style:
            요청할 widget의 형식을 지정합니다. 기본값은 flat로 설정되어 있습니다.
        :type style:
            WidgetStyle, optional

        :param scale:
            요청할 widget의 크기를 지정합니다. 반드시 0.5이상이어야 합니다. 기본값은 1.0입니다.
        :type scale:
            float, optional

        :param icon:
            요청할 widget의 아이콘을 표시할지를 지정합니다. 기본값은 False입니다.
        :type icon:
            bool, optional

        :return:
            위젯 url을 반환합니다.
        :rtype: str
        """
        if scale < 0.5:
            raise ValueError(f"scale must be greater than to 0.5, not {scale}")

        return (
            KOREANBOTS_URL
            + f"/widget/bots/{widget_type}/{bot_id}.svg?style={style}&scale={scale}&icon={icon}"
        )

    async def get_user_info(self, user_id: int):
        """
        주어진 user_id로 user의 정보를 반환합니다.

        :param user_id:
            요청할 user의 ID를 지정합니다.
        :type user_id:
            int
        """
        return await self.request("GET", f"/users/{user_id}")



from typing import Any, Dict, Type, Union


class KoreanbotsException(Exception):
    pass


class AuthorizeError(KoreanbotsException):
    pass


class HTTPException(KoreanbotsException):
    def __init__(self, code: Any, message: Union[Any, Dict[str, Any]]):
        self.status = code
        if isinstance(message, dict):
            self.status = message.get("code", self.status)
            self.error = message.get("message", "KoreanbotsException")
        else:
            self.error = message
        super().__init__(f"{self.status} {self.error}")


class BadRequest(HTTPException):
    pass


class Forbidden(HTTPException):
    pass


class NotFound(HTTPException):
    pass


ERROR_MAPPING: Dict[int, Type[HTTPException]] = {
    400: BadRequest,
    403: Forbidden,
    404: NotFound,
}



import asyncio
from asyncio.events import get_event_loop
from typing import Optional
from logging import getLogger

import aiohttp

from .decorator import strict_literal
from .http import KoreanbotsRequester
from .model import KoreanbotsBot, KoreanbotsUser
from .typing import Client, WidgetStyle, WidgetType

log = getLogger(__name__)


class Koreanbots(KoreanbotsRequester):
    """
    KoreanbotsRequester를 감싸는 클라이언트 클래스 입니다.

    :param client:
        discord.Client의 클래스입니다. 만약 필요한 경우 이 인수를 지정하세요.
    :type client:
        Optional[Client]

    :param api_key:
        API key를 지정합니다. 만약 필요한 경우 이 키를 지정하세요.
    :type api_key:
        Optional[str]

    :param session:
        aiohttp.ClientSession의 클래스입니다. 만약 필요한 경우 이 인수를 지정하세요. 지정하지 않으면 생성합니다.
    :type session:
        Optional[aiohttp.ClientSession]

    :param run_task:
        봇 정보를 갱신하는 작업을 자동으로 실행합니다. 만약 아니라면 지정하지 않습니다.
    :type run_task:
        bool

    :param include_shard_count:
        샤드 갯수를 포함할지 지정합니다. 만약 아니라면 지정하지 않습니다.
    :type include_shard_count:
        bool
    """

    def __init__(
        self,
        client: Optional[Client] = None,
        api_key: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
        run_task: bool = False,
        include_shard_count: bool = False,
    ) -> None:
        self.client = client

        # Patch discord.py client.close() method to handle session.close()
        if client:
            original_close = client.close

            async def close():
                if self.session is not None and not self.session.closed:
                    await self.session.close()
                await original_close()

            client.close = close

        self.include_shard_count = include_shard_count
        super().__init__(api_key, session)

        if run_task:
            get_event_loop().create_task(self.tasks_send_guildcount())

    async def tasks_send_guildcount(self) -> None:
        """
        길드 갯수를 서버에 전송하는 태스크 입니다.

        :raises RuntimeError:
            클라이언트를 찾을 수 없습니다.
        """

        if not self.client:
            raise RuntimeError("Client Not Found")

        await self.client.wait_until_ready()

        while not self.client.is_closed():
            kwargs = {"servers": len(self.client.guilds)}
            if self.include_shard_count:
                if self.client.shard_count:
                    kwargs.update({"shards": self.client.shard_count})
            log.info("Send")
            await self.guildcount(self.client.user.id, **kwargs)
            log.info("Complete i will sleep")
            await asyncio.sleep(1800)

    async def guildcount(self, bot_id: int, **kwargs: Optional[int]) -> None:
        """
        길드 갯수를 서버에 전송합니다.

        :param bot_id:
            요청할 bot의 ID를 지정합니다.
        :type bot_id:
            int
        """
        await self.post_update_bot_info(bot_id, **kwargs)

    async def userinfo(self, user_id: int) -> KoreanbotsUser:
        """
        유저 정보를 가져옵니다.

        :param user_id:
            요청할 유저의 ID를 지정합니다.
        :type user_id:
            int
        :return:
            유저 정보를 담고 있는 KoreanbotsUser클래스입니다.
        :rtype:
            KoreanbotsUser
        """
        return KoreanbotsUser(**await self.get_user_info(user_id))

    async def botinfo(self, bot_id: int) -> KoreanbotsBot:
        """
        봇 정보를 가져옵니다.

        :param bot_id:
            요청할 봇의 ID를 지정합니다.
        :type bot_id:
            int

        :return:
            봇 정보를 담고 있는 KoreanbotsBot클래스입니다.
        :rtype:
            KoreanbotsBot
        """
        return KoreanbotsBot(**await self.get_bot_info(bot_id))

    @strict_literal("widget_type")
    @strict_literal("style")
    async def widget(
        self,
        widget_type: WidgetType,
        bot_id: int,
        style: WidgetStyle = "flat",
        scale: float = 1.0,
        icon: bool = False,
    ) -> str:
        """
        주어진 bot_id로 widget의 url을 반환합니다.

        :param widget_type:
            요청할 widget의 타입을 지정합니다.
        :type widget_type:
            WidgetType

        :param bot_id:
            요청할 bot의 ID를 지정합니다.
        :type bot_id:
            int

        :param style:
            요청할 widget의 형식을 지정합니다. 기본값은 flat로 설정되어 있습니다.
        :type style:
            WidgetStyle, optional

        :param scale:
            요청할 widget의 크기를 지정합니다. 반드시 0.5이상이어야 합니다. 기본값은 1.0입니다.
        :type scale:
            float, optional

        :param icon:
            요청할 widget의 아이콘을 표시할지를 지정합니다. 기본값은 False입니다.
        :type icon:
            bool, optional

        :return:
            위젯 url을 반환합니다.
        :rtype: str
        """
        return await self.get_bot_widget_url(widget_type, bot_id, style, scale, icon)



from typing import Literal, TypeVar

import discord


Client = TypeVar("Client", bound=discord.Client)


WidgetType = Literal["votes", "servers", "status"]

WidgetStyle = Literal["classic", "flat"]

Category = Literal[
    "관리",
    "뮤직",
    "전적",
    "게임",
    "도박",
    "로깅",
    "빗금 명령어",
    "웹 대시보드",
    "밈",
    "레벨링",
    "유틸리티",
    "대화",
    "NSFW",
    "검색",
    "학교",
    "코로나19",
    "번역",
    "오버워치",
    "리그 오브 레전드",
    "배틀그라운드",
    "마인크래프트",
]

Status = Literal[
    "online",
    "idle",
    "dnd",
    "streaming",
    "offline",
]

State = Literal[
    "ok",
    "reported",
    "blocked",
    "private",
    "archived",
]



from typing import Any, Dict, List, Optional, Union

from .abc import KoreanbotsABC
from .typing import Category, State, Status


class BaseKoreanbots(KoreanbotsABC):
    def __init__(self, **response_data: Any) -> None:
        self._response_data = response_data

    @property
    def code(self) -> int:
        """
        상태 코드를 반환합니다.

        :return:
            상태 코드
        :rtype:
            int
        """
        return self._response_data.get("code", 0)

    @property
    def version(self) -> int:
        """
        버전을 반환합니다.

        :return:
            버전
        :rtype:
            int
        """
        return self._response_data.get("version", 0)

    @property
    def data(self) -> Dict[str, Any]:
        """
        데이터를 반환합니다.

        :return:
            데이터 딕셔너리
        :rtype:
            Dict(str, Any)
        """
        return self._response_data.get("data", {})


class KoreanbotsBot(BaseKoreanbots):
    """
    봇의 정보를 가져왔을떄 반환되는 인스턴스입니다.

    :param init_in_user:
        유저 클래스에서 초기화 되었는지 여부입니다. 기본값은 False입니다.
    :type init_in_user: bool

    """

    def __init__(
        self,
        init_in_user: bool = False,
        **response_data: Any,
    ) -> None:
        super().__init__(**response_data)
        self.init_in_user = init_in_user

    @property
    def id(self) -> Optional[str]:
        """
        아이디를 반환합니다.

        :return:
            아이디

        :rtype:
            Optional[str]
        """
        return self.data.get("id")

    @property
    def name(self) -> Optional[str]:
        """
        이름을 반환합니다.

        :return:
            이름
        :rtype:
            Optional[str]
        """
        return self.data.get("name")

    @property
    def tag(self) -> Optional[str]:
        """
        태그를 반환합니다.

        :return:
            태그
        :rtype:
            Optional[str]
        """
        return self.data.get("tag")

    @property
    def avatar(self) -> Optional[str]:
        """
        아바타를 반환합니다.


        :return:
            아바타 url
        :rtype:
            Optional[str]
        """
        return self.data.get("avatar")

    @property
    def owners(self) -> Union[List["KoreanbotsUser"], List[str]]:
        """
        소유자를 반환합니다.
        ※ init_in_user가 True인경우 소유자들의 ID를 반환합니다.

        :return:
            소유자들의 ID들을 담고 있는 리스트 또는 KoreanbotsUser 인스턴스를 담고있는 리스트
        :rtype:
            Union[List[str], List[KoreanbotsUser]]

        """
        if self.init_in_user:
            return self.response_data.get("owners", [])

        return list(
            map(
                lambda user: KoreanbotsUser(True, **user),
                self.data.get("owners", []),
            )
        )

    @property
    def flags(self) -> int:
        """
        플래그를 반환합니다.

        :return:
            플래그
        :rtype:
            int
        """
        return self.data.get("flags", 0)

    @property
    def lib(self) -> Optional[str]:
        """
        봇이 사용한 라이브러리를 반환합니다.

        :return:
            라이브러리
        :rtype:
            Optional[str]
        """
        return self.data.get("lib")

    @property
    def prefix(self) -> Optional[str]:
        """
        프리픽스를 반환합니다.

        :return:
            프리픽스
        :rtype:
            Optional[str]
        """
        return self.data.get("prefix")

    @property
    def votes(self) -> int:
        """
        투표수를 반환합니다.

        :return:
            투표수
        :rtype:
            int
        """
        return self.data.get("votes", 0)

    @property
    def servers(self) -> int:
        """
        서버수를 반환합니다.

        :return:
            서버수
        :rtype:
            int
        """
        return self.data.get("servers", 0)

    @property
    def shards(self) -> int:
        """
        샤드수를 반환합니다.

        :return:
            샤드수
        :rtype:
            int
        """
        return self.data.get("shards", 0)

    @property
    def intro(self) -> Optional[str]:
        """
        소개문구를 반환합니다.

        :return:
            소개문구
        :rtype:
            Optional[str]
        """
        return self.data.get("intro")

    @property
    def desc(self) -> Optional[str]:
        """
        설명문구를 반환합니다.

        :return:
            설명문구
        :rtype:
            Optional[str]
        """
        return self.data.get("desc")

    @property
    def web(self) -> Optional[str]:
        """
        웹사이트 주소를 반환합니다.

        :return:
            웹사이트 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("web")

    @property
    def git(self) -> Optional[str]:
        """
        깃 주소를 반환합니다.

        :return:
            깃 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("git")

    @property
    def url(self) -> Optional[str]:
        """
        주소를 반환합니다

        :return:
            주소
        :rtype:
            Optional[str]
        """
        return self.data.get("url")

    @property
    def discord(self) -> Optional[str]:
        """
        디스코드 주소를 반환합니다.

        :return:
            디스코드 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("discord")

    @property
    def category(self) -> Optional[Category]:
        """
        카테고리를 반환합니다.

        :return:
            카테고리
        :rtype:
            Optional[Category]
        """
        return self.data.get("category")

    @property
    def vanity(self) -> Optional[str]:
        """
        봇의 가상 주소를 반환합니다.

        :return:
            봇의 가상 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("vanity")

    @property
    def bg(self) -> Optional[str]:
        """
        배경 이미지 주소를 반환합니다.

        :return:
            배경 이미지 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("bg")

    @property
    def banner(self) -> Optional[str]:
        """
        배너 이미지 주소를 반환합니다.

        :return:
            배너 이미지 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("banner")

    @property
    def status(self) -> Optional[Status]:
        """
        상태를 반환합니다.

        :return:
            상태
        :rtype:
            Optional[Status]
        """
        return self.data.get("status")

    @property
    def state(self) -> Optional[State]:
        """
        Koreanbots에서의 상태를 반환합니다.
        :return:
            Koreanbots에서의 상태
        :rtype:
            Optional[State]
        """
        return self.data.get("state")


class KoreanbotsUser(BaseKoreanbots):
    """
    유저 정보를 가져왔을때 반환되는 클래스입니다.

    :param init_in_bot:
        봇 클래스에서 초기화 되었는지 여부입니다. 기본값은 False입니다.
    :type init_in_bot: bool

    """

    def __init__(self, init_in_bot: bool = False, **response_data: Any) -> None:
        super().__init__(**response_data)
        self.init_in_bot = init_in_bot

    @property
    def id(self) -> int:
        """
        ID를 반환합니다.

        :return:
            ID
        :rtype:
            int
        """
        return self.data.get("id", 0)

    @property
    def username(self) -> str:
        """
        유저 이름을 반환합니다.

        :return:
            유저 이름
        :rtype:
            str
        """
        return self.data.get("username", "")

    @property
    def tag(self) -> str:
        """
        태그를 반환합니다.

        :return:
            태그
        :rtype:
            str
        """
        return self.data.get("tag", "")

    @property
    def github(self) -> Optional[str]:
        """
        GitHub 주소를 반환합니다.

        :return:
            GitHub 주소
        :rtype:
            Optional[str]
        """
        return self.data.get("github", None)

    @property
    def flags(self) -> int:
        """
        플래그를 반환합니다.

        :return:
            플래그
        :rtype:
            int
        """
        return self.data.get("flags", 0)

    @property
    def bots(
        self,
    ) -> Union[List[KoreanbotsBot], List[str]]:
        """
        봇들을 반환합니다.
        ※ init_in_bot가 True인경우 봇들의 ID를 반환합니다.

        :return:
            봇들의 ID들을 담고 있는 리스트 또는 KoreanbotsUser 인스턴스를 담고있는 리스트
        :rtype:
            Union[List[str], List[KoreanbotsUser]]

        """
        if self.init_in_bot:
            return self.response_data.get("bots", [])

        return list(
            map(
                lambda bot: KoreanbotsBot(True, **bot),
                self.data.get("bots", []),
            )
        )



import abc
from typing import Any, Dict


class KoreanbotsABC(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, **response_data: Any) -> None:
        self.response_data = response_data

    @abc.abstractproperty
    def code(self) -> int:
        return self.response_data.get("code", 0)

    @abc.abstractproperty
    def version(self) -> int:
        return self.response_data.get("version", 0)

    @abc.abstractproperty
    def data(self) -> Dict[str, Any]:
        return self.response_data.get("data", {})



# -*- coding: utf-8 -*-

from setuptools import setup
import os
import koreanbots

version = koreanbots.__version__

path = os.path.dirname(os.path.realpath(__file__)).replace("\\", "/")

requirements = []
with open(f"{path}/requirements.txt", encoding="UTF8") as f:
    requirements = f.read().splitlines()

if not version:
    raise RuntimeError("version is not defined")

readme = ""
with open(f"{path}/README.md", encoding="UTF8") as f:
    readme = f.read()

setup(
    name="koreanbots",
    author="Koreanbots",
    url="https://github.com/koreanbots/py-sdk",
    project_urls={
        "Homepage": "https://koreanbots.dev/",
        "Source": "https://github.com/koreanbots/py-sdk",
        "Tracker": "https://github.com/koreanbots/py-sdk/issues",
    },
    version=version,
    packages=["koreanbots"],
    license="MIT",
    description="Official SDK for Koreanbots.",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import functools
import inspect
from typing import Any, Literal, get_args


def strict_literal(argument_name: str):
    def decorator(f: Any):
        @functools.wraps(f)
        async def decorated_function(*args: Any, **kwargs: Any):
            # First get about func args
            full_arg_spec = inspect.getfullargspec(f)
            # Get annotation
            arg_annoration = full_arg_spec.annotations[argument_name]
            # Check annotation is Lireral
            if arg_annoration.__origin__ is Literal:
                # Literal -> list
                literal_list = list(get_args(arg_annoration))
                # Get index
                arg_index = full_arg_spec.args.index(argument_name)
                # Handle arguments
                if arg_index < len(args) and args[arg_index] not in literal_list:
                    raise ValueError(
                        f"Arguments do not match. Expected: {literal_list}"
                    )
                # Handle keyword arguments
                elif (
                    kwargs.get(argument_name)
                    and kwargs[argument_name] not in literal_list
                ):
                    if kwargs[argument_name] not in literal_list:
                        raise ValueError(
                            f"Arguments do not match. Expected: {literal_list}"
                        )

            return await f(*args, **kwargs)

        return decorated_function

    return decorator



from .client import Koreanbots
from .http import KoreanbotsRequester
from .errors import *
from .model import KoreanbotsBot, KoreanbotsUser

from typing import Literal, NamedTuple


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int


versioninfo = VersionInfo(1, 0, 0, "final", 0)

__version__ = "1.0.0"
